# Name:         warnings.py
# Purpose:      Definitions of nansat warnings
# Authors:      Morten W. Hansen
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html

class NansatFutureWarning(Warning):
    pass
